
import pandas as pd
from twizzle import Twizzle


class AnalysisDataGenerator(object):
    """Generator for analysis data in pandas format
    """

    def __init__(self, sDBPath):
        """Constructor of the AnalysisDataGenerator class

        Note:
            Please define the `DB_PATH` in the config.py or pass the path of the SQLite 
            as parameter
        Args:
            sDBPath (str): Path to the SQLite database.
        """
        if sDBPath is None:
            raise Exception("Path to SQL-Database has to be defined")
        tw = Twizzle(sDBPath)
        dfChallenges = pd.DataFrame(tw.get_challenges())
        if dfChallenges.empty:
            raise Exception("currently there are no challenges defined yet")
        dfTests = pd.DataFrame(tw.get_tests())
        if dfTests.empty:
            raise Exception("currently no test have been run yet")

        dfChallenges = dfChallenges.drop(
            labels=["originalImages", "comparativeImages", "targetDecisions"], axis=1)
        dfTests = pd.merge(dfTests, dfChallenges, how="inner",
                           left_on="challenge", right_on="challenge")
        self.dataframe = dfTests

    def get_pandas_dataframe(self):
        """ get concatenated analysis data as pandas dataframe
        Returns:
            Pandas Dataframe concatenating all challenges with all tests and parameters
        """
        return self.dataframe

    def save_pandas_dataframe_to_file(self, sPathToFile):
        """ save concatenated analysis data as pandas dataframe to CSV file
        """
        if not sPathToFile:
            raise Exception("path to file not specified")
        self.dataframe.to_csv(sPathToFile)



#!/usr/bin/env python3


"""
This module defines common functions to calculate the deviation between two hashes
"""


import numpy as np


def hamming_distance(array1, array2):
    if (array1.size != array2.size):
        raise Exception("Arrays have to have the same size")
    hamming = array1 != array2
    return np.count_nonzero(hamming) / hamming.size



from twizzle.twizzle import Twizzle
from twizzle.analysis_data_generator import AnalysisDataGenerator
from twizzle.test_runner import TestRunner
import twizzle.attacks_preset as attacks_preset
import twizzle.utils as utils
import twizzle.deviation_presets as deviation_presets
import twizzle.hashalgos_preset as hashalgos_preset



"""
This module defines common perceptual image hashing algorithms that can be used as an example
"""

import numpy
import cv2


def __resize_image_downscale(aInputImage, lImageWidth, lImageHeight):
    """resizing an image to a given size"""
    return cv2.resize(aInputImage, (lImageWidth, lImageHeight), interpolation=cv2.INTER_AREA)


def __convert_image_to_grayscale(aInputImage):
    """converting an image to grayscale"""
    return cv2.cvtColor(aInputImage, cv2.COLOR_BGR2GRAY)


def average_hash(image, hash_size=8):
    """
    Average Hash computation

    Implementation follows http://www.hackerfactor.com/blog/index.php?/archives/432-Looks-Like-It.html

    Step by step explanation: https://www.safaribooksonline.com/blog/2013/11/26/image-hashing-with-python/
    """
    if hash_size < 0:
        raise ValueError("Hash size must be positive")

    # reduce size and complexity, then covert to grayscale
    # image = image.convert("L").resize((hash_size, hash_size),
    # Image.ANTIALIAS)
    image = __convert_image_to_grayscale(image)
    pixels = __resize_image_downscale(image, hash_size, hash_size)

    # find average pixel value; 'pixels' is an array of the pixel values,
    # ranging from 0 (black) to 255 (white)
    # pixels = numpy.array(image.getdata()).reshape((hash_size, hash_size))
    avg = pixels.mean()

    # create string of bits
    diff = pixels > avg
    # make a hash
    return diff.flatten()


def dhash(image, hash_size=8):
    """
    Difference Hash computation.

    following http://www.hackerfactor.com/blog/index.php?/archives/529-Kind-of-Like-That.html

    computes differences horizontally

    """
    if hash_size < 0:
        raise ValueError("Hash size must be positive")

    image = __convert_image_to_grayscale(image)
    pixels = __resize_image_downscale(
        image, hash_size + 1, hash_size)
    # compute differences between columns
    diff = pixels[:, 1:] > pixels[:, :-1]
    return diff.flatten()



from twizzle import Twizzle
from multiprocessing.pool import ThreadPool
from threading import Lock


class TestRunner(object):
    """ TestRunner - creates a multi threaded environment for running tests
    """

    def __init__(self, sDBPath, lNrOfThreads=2):
        """Constructor of a TestRunner class

        Note:
            Please define the `DB_PATH` in the config.py or pass the path of the SQLite 
            as parameter
        Args:
            sDBPath (str): Path to the SQLite database.
            lNrOfThreads (int): number of threads to use for the tests
        """
        if sDBPath is None:
            raise Exception("Path to SQL-Database has to be defined")
        if lNrOfThreads <= 0:
            raise Exception("lNrOfThreads has to be grater then 0")
        self.tw = Twizzle(sDBPath)
        self.oPool = ThreadPool(processes=lNrOfThreads)
        self.aTaskPoolThreads = []
        self.lock = Lock()

    def run_test_async(self, sChallengeName, fnCallback, dicCallbackParameters={}):
        """add test run to threadpool

        Args:
            sChallengeName (str): name of the challenge that should be tested
            fnCallback (function): test wrapper function that should be called 
            dicCallbackParameters (:obj:): Dictionary of parameters for  fnCallback

        Returns:
            None
        """
        pThread = self.oPool.apply_async(
            self.tw.run_test, (sChallengeName, fnCallback, dicCallbackParameters))
        self.aTaskPoolThreads.append(pThread)

    def wait_till_tests_finished(self):
        """block execution till all threads are done"""
        # catch threads ready
        for pThread in self.aTaskPoolThreads:
            dicTest = pThread.get()
            self.tw.save_test_threadsafe(dicTest, self.lock)

    def get_tests(self):
        """get all tests defined"""
        return self.tw.get_tests()



#!/usr/bin/env python3

"""
This module implements many common image manipulation attacks.
Most of them can be used in the challenge creator.
"""


import numpy as np

import cv2
# from scipy import ndimage
from skimage import exposure
import blend_modes
from skimage import util
from skimage import img_as_float
from skimage import img_as_ubyte
import math


def scale(aInputImage, lScalefactorX=1, lScaleFactorY=1):
    """scaling an image by a given scale factor"""
    if (lScalefactorX <= 1 and lScaleFactorY <= 1):
        fnInterpolation = cv2.INTER_AREA
    else:
        fnInterpolation = cv2.INTER_CUBIC
    return cv2.resize(aInputImage, None, fx=lScalefactorX, fy=lScaleFactorY, interpolation=fnInterpolation)


def rotation_cropped(aInputImage, dRotationAngle=25):
    """ perform a cropped rotation that crops the rotated image to fill the whole canvas without borders"""
    def rotatedRectWithMaxArea(w, h, angle):
        # thanks to coproc
        # (https://stackoverflow.com/questions/16702966/rotate-image-and-crop-out-black-borders)
        """
        Given a rectangle of size wxh that has been rotated by 'angle' (in
        radians), computes the width and height of the largest possible
        axis-aligned rectangle (maximal area) within the rotated rectangle.
        """
        if w <= 0 or h <= 0:
            return 0, 0

        width_is_longer = w >= h
        side_long, side_short = (w, h) if width_is_longer else (h, w)

        # since the solutions for angle, -angle and 180-angle are all the same,
        # if suffices to look at the first quadrant and the absolute values of
        # sin,cos:
        sin_a, cos_a = abs(math.sin(angle)), abs(math.cos(angle))
        if side_short <= 2. * sin_a * cos_a * side_long or abs(sin_a - cos_a) < 1e-10:
            # half constrained case: two crop corners touch the longer side,
            # the other two corners are on the mid-line parallel to the longer
            # line
            x = 0.5 * side_short
            wr, hr = (x / sin_a, x /
                      cos_a) if width_is_longer else (x / cos_a, x / sin_a)
        else:
            # fully constrained case: crop touches all 4 sides
            cos_2a = cos_a * cos_a - sin_a * sin_a
            wr, hr = (w * cos_a - h * sin_a) / \
                cos_2a, (h * cos_a - w * sin_a) / cos_2a

        return int(wr), int(hr)

    lOriginalY = aInputImage.shape[0]
    lOriginalX = aInputImage.shape[1]
    # rotate about the middle point
    M = cv2.getRotationMatrix2D(
        center=((lOriginalX - 1) / 2, (lOriginalY - 1) / 2), angle=dRotationAngle, scale=1.0)
    lNewX, lNewY = rotatedRectWithMaxArea(
        lOriginalX, lOriginalY, math.radians(dRotationAngle))
    (tx, ty) = ((lNewX - lOriginalX) / 2, (lNewY - lOriginalY) / 2)
    # third column of matrix holds translation, which takes effect after
    # rotation.
    M[0, 2] += tx
    M[1, 2] += ty
    return cv2.warpAffine(aInputImage, M, dsize=(lNewX, lNewY))


def rotation(aInputImage,  dRotationAngle=30, bFit=True, tpBorderValue=(0, 0, 0)):
    # thanks to Lars Schillingmann (https://stackoverflow.com/questions/22041699/rotate-an-image-without-cropping-in-opencv-in-c)
    # note: numpy uses (y,x) convention but most OpenCV functions use (x,y)
    lOriginalY = aInputImage.shape[0]
    lOriginalX = aInputImage.shape[1]
    # rotate about the middle point
    M = cv2.getRotationMatrix2D(
        center=((lOriginalX - 1) / 2, (lOriginalY - 1) / 2), angle=dRotationAngle, scale=1.0)

    if bFit:
        # calculate new image size
        lNewX, lNewY = lOriginalX, lOriginalY
        # include this if you want to prevent corners being cut off
        dAngle = np.deg2rad(dRotationAngle)
        lNewX, lNewY = (abs(np.sin(dAngle) * lNewY) + abs(np.cos(dAngle) * lNewX),
                        abs(np.sin(dAngle) * lNewX) + abs(np.cos(dAngle) * lNewY))

        (tx, ty) = ((lNewX - lOriginalX) / 2, (lNewY - lOriginalY) / 2)
        # third column of matrix holds translation, which takes effect after
        # rotation.
        M[0, 2] += tx
        M[1, 2] += ty

        aRotatedImg = cv2.warpAffine(
            aInputImage, M, dsize=(int(lNewX), int(lNewY)), borderValue=tpBorderValue)
    else:
        aRotatedImg = cv2.warpAffine(
            aInputImage, M, dsize=(lOriginalX, lOriginalY), borderValue=tpBorderValue)
    return aRotatedImg


def flip(aInputImage, bVertical=False):
    if(bVertical):
        return np.flipud(aInputImage)
    else:
        return np.fliplr(aInputImage)


def perspective_transformation(aInputImage, tl=(0, 0), tr=(0, 0), bl=(0, 0), br=(0, 0), bResize=True, tpBorderValue=(0, 0, 0)):
    """transform an image, move the four edge point, for every point you can give (x,y) where x and y can be positive or negative"""

    lImgHeight = aInputImage.shape[0]
    lImgWidth = aInputImage.shape[1]
    rect = np.array([
        [0, 0],
        [lImgWidth - 1, 0],
        [lImgWidth - 1, lImgHeight - 1],
        [0, lImgHeight - 1]
    ], dtype="float32")

    dst = np.array([
        [0 + tl[0], 0 + tl[1]],
        [lImgWidth - 1 + tr[0], 0 + tr[1]],
        [lImgWidth - 1 + br[0], lImgHeight - 1 + br[1]],
        [0 + bl[0], lImgHeight - 1 + bl[1]]
    ], dtype="float32")
    M = cv2.getPerspectiveTransform(rect, dst)

    lDeltaWidth = 0
    lDeltaHeight = 0
    if bResize:
        lXTranslation = -min(tl[0], bl[0])
        lYTranslation = -min(tl[1], tr[1])
        lDeltaWidth = lXTranslation + max(br[0], tr[0])
        lDeltaHeight = lYTranslation + max(bl[1], br[1])
        aTranslation = np.array([
            [1, 0, lXTranslation],
            [0, 1, lYTranslation],
            [0, 0, 1]
        ], dtype="float32")
        M = np.dot(aTranslation, M)

    return cv2.warpPerspective(aInputImage, M, (lImgWidth + lDeltaWidth, lImgHeight + lDeltaHeight), cv2.INTER_LINEAR, cv2.BORDER_CONSTANT, borderValue=tpBorderValue)


def crop(aInputImage, tpSlice):
    """crops an image, x pixel according to a slice tuple (top, left, buttom, right)"""
    t, l, b, r = tpSlice
    lImgHeight = aInputImage.shape[0]
    lImgWidth = aInputImage.shape[1]
    if t + b >= lImgHeight or l + r >= lImgWidth:
        raise Exception("Crop area can not be bigger that the image.")
    return aInputImage[t:lImgHeight - b, l:lImgWidth - r]


def crop_percentage(aInputImage, tpSlice):
    """crops an image, x percent according to a slice tuple (top, left, buttom, right)"""
    tp, lp, bp, rp = tpSlice
    if tp + bp >= 1 or lp + rp >= 1:
        raise Exception("Crop area can not be bigger that the image.")
    lImgHeight = aInputImage.shape[0]
    lImgWidth = aInputImage.shape[1]
    t = int(np.floor(tp * lImgHeight))
    b = int(np.floor(bp * lImgHeight))
    l = int(np.floor(lp * lImgWidth))
    r = int(np.floor(rp * lImgWidth))

    return crop(aInputImage, (t, l, b, r))

# def colored_noise(aInputImage, mean=5, stddev=128):
# return cv2.randn(np.copy(aInputImage), (mean, mean, mean), (stddev,
# stddev, stddev)) + aInputImage


# def noise_gauss(aInputImage, dSigma=0.1):
#     """ applies gaussian noise to an image """
#     row, col, ch = aInputImage.shape
#     lMean = 0
#     aGauss = np.random.normal(lMean, dSigma, (row, col, ch))
#     aGauss = aGauss.reshape(row, col, ch)
#     return (aInputImage + aGauss).astype(np.dtype(np.uint8))


def brightness(aInputImage, lBrightness=150):
    aBrightedImage = aInputImage.astype(np.int16) + lBrightness
    aBrightedImage[aBrightedImage > 255] = 255
    aBrightedImage[aBrightedImage < 0] = 0
    return aBrightedImage.astype(np.uint8)


def contrast(aInputImage, lContrast=50):
    """ changes the contrast of a given image by histogram manipulation
        takes values between [-128, 128] as contrast factor
    """
    if abs(lContrast) > 128:
        raise Exception("contrast value have to be in range [-128, 128]")
    if lContrast >= 0:
        return exposure.rescale_intensity(aInputImage, in_range=(0 + lContrast, 255 - lContrast))
    else:
        return exposure.rescale_intensity(aInputImage, out_range=(0 + abs(lContrast), 255 - abs(lContrast)))


# def gaussian_filter(aInputImage, dSigma=2.0):
#     """ applies a gaussian filter on a given image """
#     return ndimage.gaussian_filter(aInputImage, dSigma)


# def median_filter(aInputImage, dSigma=3.0):
#     return ndimage.median_filter(aInputImage, dSigma)


def gaussian_filter(aInputImage, lSigma=5):
    """apply a gaussian filter on a given image
       the size of the kernel ist calculated from sigma
    """
    return cv2.GaussianBlur(aInputImage, (0, 0), lSigma)


def median_filter(aInputImage, lKernelSize=5):
    """apply a median filter on a given image"""
    return cv2.medianBlur(aInputImage, lKernelSize)


def add_pattern(aInputImage, aPatternImage, dAlphaPattern=0.2):
    return cv2.addWeighted(aInputImage, 1.0 - dAlphaPattern, aPatternImage, dAlphaPattern, 0.0)


def blend_image(aInputImage, aBGImage, dOpacity=1.0, sBlendMode="mul"):
    """ blending pattern to image"""
    if aInputImage.shape != aBGImage.shape:
        raise Exception(
            "Image and BackgrundImage have to be the same Size and color Model")
    aInputImageAlpha = __add_alpha_channel(aInputImage)
    aBGImageAlpha = __add_alpha_channel(aBGImage)

    dicBlendModes = {
        "sl": blend_modes.soft_light,
        "lo": blend_modes.lighten_only,
        "do": blend_modes.dodge,
        "ad": blend_modes.addition,
        "dar": blend_modes.darken_only,
        "mul": blend_modes.multiply,
        "hl": blend_modes.hard_light,
        "dif": blend_modes.difference,
        "sub": blend_modes.subtract,
        "gre": blend_modes.grain_extract,
        "grm": blend_modes.grain_merge,
        "div": blend_modes.divide}
    if sBlendMode in dicBlendModes.keys():
        fnBlendMode = dicBlendModes[sBlendMode]
    else:
        fnBlendMode = dicBlendModes["mul"]

    aBlendedImageAlpha = fnBlendMode(
        aBGImageAlpha.astype(float), aInputImageAlpha.astype(float), dOpacity)
    aBlendedImage = __remove_alpha_channel(aBlendedImageAlpha)
    return np.uint8(aBlendedImage)


def blend_pattern(aInputImage, aPatternImage=None, dOpacity=1.0, sBlendMode="mul"):
    """ stretches the pattern image to the size of the image and blends them """
    if aPatternImage is None:
        raise Exception("no image was given as pattern")
    aPatternRightSize = __resize_pattern(aInputImage, aPatternImage)
    return blend_image(aInputImage, aPatternRightSize, dOpacity, sBlendMode)


def __resize_pattern(aInputImage, aPatternImage):
    """ resize a pattern image to the whished size"""
    lTargetHeight, lTargetWidth = aInputImage.shape[:2]
    lPatternHeigh, lPatternWidth = aPatternImage.shape[:2]
    # if pattern is already bigger than necessary
    if lTargetHeight < lPatternHeigh and lTargetWidth < lPatternWidth:
        return aPatternImage[:lTargetHeight, :lTargetWidth]
    # calculate multiplier
    lPatternHeighMultiplier = math.ceil(lTargetHeight / lPatternHeigh)
    lPatternWidthMultiplier = math.ceil(lTargetWidth / lPatternWidth)
    lPatternColorMultiplier = 0 if len(aPatternImage.shape) != 3 else 1
    # multiply the pattern
    aMultipliedPattern = np.tile(
        aPatternImage, (lPatternHeighMultiplier, lPatternWidth, lPatternColorMultiplier))
    return aMultipliedPattern[:lTargetHeight, :lTargetWidth]


def __add_alpha_channel(aInputImage):
    """ adds an alpha chanel to the image """
    aBChannel, aGChannel, aRChannel = cv2.split(aInputImage)
    aAlphaChannel = np.ones(aBChannel.shape, dtype=aBChannel.dtype)
    return cv2.merge((aBChannel, aGChannel, aRChannel, aAlphaChannel * 255))


def __remove_alpha_channel(aInputImage):
    """ removes the alpha channel from image """
    return aInputImage[:, :, :3]


def gamma_adjustment(aInputImage, dGamma=1.0, dGain=1.0):
    """ adapts the gamma exposure """
    return exposure.adjust_gamma(aInputImage, dGamma, dGain)


def shift_vertical(aInputImage, lPixles=10, bFillArea=False, tpBorderValue=(0, 0, 0)):
    """ shifts an image by x pixles vertically """
    aShiftedImage = np.roll(aInputImage, lPixles, 0)
    if bFillArea:
        if lPixles >= 0:
            aShiftedImage[:lPixles, :] = tpBorderValue
        else:
            aShiftedImage[lPixles:, :] = tpBorderValue
    return aShiftedImage


def shift_horizontal(aInputImage, lPixles=10, bFillArea=False, tpBorderValue=(0, 0, 0)):
    """ shifts an image by x pixles horizontally """
    aShiftedImage = np.roll(aInputImage, lPixles, 1)
    if bFillArea:
        if lPixles >= 0:
            aShiftedImage[:, :lPixles] = tpBorderValue
        else:
            aShiftedImage[:, lPixles:] = tpBorderValue
    return aShiftedImage


def jpeg_compression(aInputImage, lJPEGQuality=95):
    """ encodes an image as jpeg with a given quality """
    r, aBuffer = cv2.imencode(".jpeg", aInputImage,
                              (cv2.IMWRITE_JPEG_QUALITY, lJPEGQuality))
    return cv2.imdecode(aBuffer, -1)


def speckle_noise(aInputImage, dSigma=0.001):
    """ adds speckle noise """
    image = img_as_float(aInputImage)
    image = util.random_noise(aInputImage, 'speckle', mean=0, var=dSigma)
    return img_as_ubyte(image)


def salt_and_pepper_noise(aInputImage, dAmount=0.001, dProportion=0.5):
    """ adds salt and pepper noise """
    image = img_as_float(aInputImage)
    image = util.random_noise(
        aInputImage, 's&p', amount=dAmount, salt_vs_pepper=dProportion)
    return img_as_ubyte(image)


def gauss_noise(aInputImage, dSigma=0.1):
    """adds gaussian noise with skimage """
    image = img_as_float(aInputImage)
    image = util.random_noise(image, 'gaussian', mean=0, var=dSigma)
    return img_as_ubyte(image)


def position_watermarking(aInputImage, aWatermarkImage, position='top_left', dOpacity=1.0, sBlendMode="mul"):
    """adds Watermark on 1 of 5 positions: top_left, bottom_left, bottom_right, top_right, middle"""
    lTargetHeight, lTargetWidth = aInputImage.shape[0], aInputImage.shape[1]
    aBlankImage = np.zeros((lTargetHeight, lTargetWidth, 3), np.uint8)
    aBlankImage[:, :] = (255, 255, 255)
    lWatermarkHeigh, lWatermarkWidth = aWatermarkImage.shape[:2]
    if position == 'top_left':
        aBlankImage[0:lWatermarkHeigh, 0:lWatermarkWidth] = aWatermarkImage
    elif position == 'bottom_left':
        aBlankImage[(lTargetHeight - lWatermarkHeigh):lTargetHeight,
                    0:lWatermarkWidth] = aWatermarkImage
    elif position == 'bottom_right':
        aBlankImage[(lTargetHeight - lWatermarkHeigh):lTargetHeight,
                    (lTargetWidth - lWatermarkWidth):lTargetWidth] = aWatermarkImage
    elif position == 'top_right':
        aBlankImage[0:lWatermarkHeigh,
                    (lTargetWidth - lWatermarkWidth):lTargetWidth] = aWatermarkImage
    elif position == 'middle':
        if lTargetHeight % 2 == 0 and lTargetWidth % 2 == 0:
            aBlankImage = cv2.resize(
                aBlankImage, (lTargetHeight + 1, lTargetWidth + 1))
        elif lTargetHeight % 2 == 0:
            aBlankImage = cv2.resize(
                aBlankImage, (lTargetHeight + 1, lTargetWidth))
        elif lTargetWidth % 2 == 0:
            aBlankImage = cv2.resize(
                aBlankImage, (lTargetHeight, lTargetWidth + 1))
        if lWatermarkHeigh % 2 == 0 and lWatermarkWidth % 2 == 0:
            aWatermarkImage = cv2.resize(
                aWatermarkImage, (lWatermarkHeigh + 1, lWatermarkWidth + 1))
        elif lWatermarkHeigh % 2 == 0:
            aWatermarkImage = cv2.resize(
                aWatermarkImage, (lWatermarkHeigh + 1, lWatermarkWidth))
        elif lWatermarkWidth % 2 == 0:
            aWatermarkImage = cv2.resize(
                aWatermarkImage, (lWatermarkHeigh, lWatermarkWidth + 1))
        (lCenterImageX, lCenterImageY) = (
            (aBlankImage.shape[0] + 1) / 2, (aBlankImage.shape[1] + 1) / 2)
        (lCenterWatermarkX, lCenterWatermarkY) = (
            (aWatermarkImage.shape[0] + 1) / 2, (aWatermarkImage.shape[1] + 1) / 2)
        X1 = int(lCenterImageX - (lCenterWatermarkX))
        Y1 = int(lCenterImageY - (lCenterWatermarkY))
        aBlankImage[X1:(X1 + aWatermarkImage.shape[0]),
                    Y1:(Y1 + aWatermarkImage.shape[1])] = aWatermarkImage
        aBlankImage = cv2.resize(aBlankImage, (lTargetWidth, lTargetHeight))
    return blend_image(aInputImage, aBlankImage, dOpacity, sBlendMode)



#!/usr/bin/env python3


""" This module defines many common utility functions used in the whole project
"""


import cv2
import os
import string


def escape_home_in_path(sPath):
    return os.path.expanduser(sPath)


def save_image(aImage, sPathToImage):
    """saves a given image to the given location"""
    cv2.imwrite(escape_home_in_path(sPathToImage), aImage)


def load_image(sPathToImage):
    """load an image from disk"""
    return cv2.imread(escape_home_in_path(sPathToImage))


def create_path(sPath):
    """ create a path if it is not existend yet"""
    sPathEscape = os.path.expanduser(sPath)
    try:
        if not os.path.exists(sPathEscape):
            os.makedirs(sPathEscape)
        return sPathEscape
    except:
        print("some error occurred while creating dir %s" % (sPathEscape))


def check_if_path_exists(sPath):
    """ returns Ture if a given path is existend """
    sPathEscape = os.path.expanduser(sPath)
    return os.path.exists(sPathEscape)


def check_if_file_exists(sPath):
    """ returns Ture if a given path is a file and is existend """
    sPathEscape = os.path.expanduser(sPath)
    return os.path.isfile(sPathEscape)


def list_all_images_in_directory(sDirectoryPath):
    """ returns as list of all images in a given subfolder """
    sDirectoryPathEscaped = escape_home_in_path(sDirectoryPath)
    if not check_if_path_exists(sDirectoryPathEscaped):
        raise Exception("path %s does not exist" % sDirectoryPathEscaped)
    if not sDirectoryPathEscaped.endswith("/"):
        sDirectoryPathEscaped += "/"
    aFiles = next(os.walk(sDirectoryPathEscaped))[2]
    # filter image files
    return [sDirectoryPathEscaped +
            m for m in aFiles if m.lower().endswith((".png", ".bmp", ".jpg", ".jpeg", ".tiff"))]


def format_filename(s):
    """
        Thanks to Sean Hammond (https://github.com/seanh)
        Take a string and return a valid filename constructed from the string.
        Uses a whitelist approach: any characters not present in valid_chars are
        removed. Also spaces are replaced with underscores.

        Note: this method may produce invalid filenames such as ``, `.` or `..`
        When I use this method I prepend a date string like '2009_01_15_19_46_32_'
        and append a file extension like '.txt', so I avoid the potential of using
        an invalid filename.

        """
    valid_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)
    filename = ''.join(c for c in s if c in valid_chars)
    filename = filename.replace(' ', '_')  # I don't like spaces in filenames.
    return filename


def get_filename_without_extension(sPathToFile):
    """
    returns the name of a file without fileextension and path
    """
    return os.path.splitext(os.path.basename(sPathToFile))[0]



#!/usr/bin/env python3

from sqlitedict import SqliteDict
import numpy as np

DB_CHALLENGES_KEY = 'challenges'
DB_TESTS_KEY = 'tests'


class Twizzle(object):
    """Twizzle multi purpose benchmarking system -- base class
    """

    def __init__(self, sDBPath):
        """Constructor of the Twizzle class

        Note:
            Please pass the path of the SQLite 
            as parameter
        Args:
            sDBPath (str): Path to the SQLite database.
        """
        if sDBPath is None:
            raise Exception("Path to SQL-Database has to be defined")
        self._db = SqliteDict(sDBPath)

    def add_challenge(self, sName, aOriginalObjects, aComparativeObjects, aTargetDecisions, dicMetadata={}):
        """Adds a challenge under the given name to the database

        Note:
            The three lists describe a table of the following format:

            | Original object | Comparative object | target decision |
            |----------------|-------------------|-----------------|
            | Img1.png       | Img1_scaled.png   | True            |
            | Img2.png       | Img2_brighter.png | True            |
            | Img2.png       | Img9.png          | False           |


        Args:
            sName (str): the name of the challenge.
            aOriginalImages (:obj:`list` of :obj:`str`): List of paths of the original images
            aComparativeImages (:obj:`list` of :obj:`str`): List of paths of the images that should be compared to 
                                                            the original image at the same position in the list
            aTargetDecisions (:obj:`list` of :obj:`bool`): List of boolean defining whether the images linked in aOriginalImages
                                                            and aComparativeImages beeing at the same position in the list are
                                                            the same (True) or not (False)
            dicMetadata (:obj:): an object defining metadata for the challenge like what printer was used or what kind of attack 
                                using which parameters was performed

        Returns:
            None
        """

        # catch wrong parameters
        if (not sName) or (aOriginalObjects is None) or (aComparativeObjects is None) or (aTargetDecisions is None):
            raise Exception("Parameters can not be None.")
        if not(len(aOriginalObjects) == len(aComparativeObjects) == len(aTargetDecisions)):
            raise Exception(
                "Image sets and target decisions have to have the same amount of entries.")
        if not (all(isinstance(x, str) for x in aOriginalObjects) and all(isinstance(x, str) for x in aComparativeObjects)):
            raise Exception(
                "All objects have to be defined as path given as string.")
        if (not all(isinstance(x, bool) for x in aTargetDecisions)) and not isinstance(aTargetDecisions, np.ndarray) and not (aTargetDecisions.dtype == np.dtype("bool")):
            raise Exception("The target decisions have to be boolean only.")

        # get current challenges from database
        aChallenges = self._db.get(DB_CHALLENGES_KEY, [])

        # test whether name was used before
        aChallengesSameName = [
            ch for ch in aChallenges if ch["challenge"] == sName]
        if len(aChallengesSameName) != 0:
            raise Exception(
                "Challenge name %s is already in use. Define an other one. Aborting." % sName)

        # append new challenge
        dicChallenge = {"challenge": sName, "originalObjects": aOriginalObjects,
                        "comparativeObjects": aComparativeObjects, "targetDecisions": aTargetDecisions}
        # adding additional information if given
        if dicMetadata:
            dicChallenge = {**dicMetadata, **dicChallenge}
        aChallenges.append(dicChallenge)
        self._db[DB_CHALLENGES_KEY] = aChallenges
        self._db.commit()

    def del_challenge(self, sName):
        """ deletes an existing challenge by its name

        Args:
            sName (str): the name of the challenge to be deleted

        Returns:
            None
        """

        # get current challenges from database
        aChallenges = self._db.get(DB_CHALLENGES_KEY, [])
        aMatches = [ch for ch in aChallenges if ch["challenge"] == sName]
        if len(aMatches) == 0:
            raise Exception("No challenge named %s found." % sName)

        # remove element
        aChallenges.remove(aMatches[0])

        # save new db
        self._db[DB_CHALLENGES_KEY] = aChallenges
        self._db.commit()

    def get_challenges(self):
        """ getting a list of all defined challenges

        Returns:
            :obj:`list` of :obj:: `obj`:  List of all defined challenges
        """
        return self._db.get(DB_CHALLENGES_KEY, [])

    def get_challenge(self, sChallengeName):
        """ getting a single challenge object

        Args:
            sChallengeName (str): the name of the challenge to get

          Returns:
            :obj:: `obj`:  Object defining the challenge having the name sChallengeName
        """
        aChallenges = self._db.get(DB_CHALLENGES_KEY, [])
        aMatches = [ch for ch in aChallenges if ch["challenge"]
                    == sChallengeName]
        if len(aMatches) == 0:
            raise Exception("No challenge with name %s found." %
                            sChallengeName)
        return aMatches[0]

    def clear_challenges(self):
        """ clears all challenge entries from the database """
        self._db[DB_CHALLENGES_KEY] = []
        self._db.commit()

    def run_test(self, sChallengeName, fnCallback, dicCallbackParameters={}, autosave_to_db=False):
        """ run single challenge as test using given callback function and optional params

        Note:
            fnCallback has to fullfill following specifications

            Parameters:
            fnCallback(aOriginalObjects, aComparativeObjects, **dicCallbackParameters)
            - aOriginalObjects: list of strings describing paths to original objects
            - aComparativeObjects: list of strings describing paths to comparative objects
            ... arbitrary number of further parameters

            Returns:
            aDecisions, dicAdditionalInformation = fnCallback(...)
            - aDecisions: list of boolean decisions describing wether the algorithm has decided that the original object 
                          and the comparative objects are the same (True) or not (False)
            - dicAdditionalInformation: the algorithm can supply additional information that can be used in the evaluation 
                                        later on to compare different settings


        Args:
            sChallengeName (str): the challenge that should be executed
            fnCallback (function): Pointer to wrapper-function that tests a challenge on a specific algorithm
                                    and makes decisions whether the objects are the same or not depending on its decision algorithm
            dicCallbackParameters (:obj:): Dictionary defining parameters for the function in fnCallback

        Returns:
            dicTest: dictionary of test results that can be saved to db
        """
        if not(sChallengeName) or not(fnCallback):
            raise Exception("Parameters are not allowed to be None.")

        dicChallenge = self.get_challenge(sChallengeName)
        sChallengeName = dicChallenge["challenge"]
        aOriginalObjects = dicChallenge["originalObjects"]
        aComparativeObjects = dicChallenge["comparativeObjects"]
        aTargetDecisions = dicChallenge["targetDecisions"]

        # run challenge
        aDecisions, dicAdditionalInformation = fnCallback(
            aOriginalObjects, aComparativeObjects, **dicCallbackParameters)

        # check if site of decisions is right
        if len(aDecisions) != len(aTargetDecisions):
            raise Exception(
                "Array of Decisions is not the same size as given set of objects. Aborting.")

        # calculate errorate
        lTestsetSize = len(aTargetDecisions)
        lErrors = np.count_nonzero(
            np.array(aDecisions) != np.array(aTargetDecisions))
        dErrorRate = lErrors / lTestsetSize

        lTP = np.sum(np.logical_and(
            aDecisions == True, aTargetDecisions == True))
        lTN = np.sum(np.logical_and(
            aDecisions == False, aTargetDecisions == False))
        lFP = np.sum(np.logical_and(
            aDecisions == True, aTargetDecisions == False))
        lFN = np.sum(np.logical_and(
            aDecisions == False, aTargetDecisions == True))
        dAccuracy = (lTP+lTN)/(lTP+lTN+lFP+lFN)
        dPrecision = lTP/(lTP+lFP)
        dRecall = lTP/(lTP+lFN)
        dF1score = 2*((dPrecision*dRecall)/(dPrecision+dRecall))
        dFAR = lFP/(lFP + lTN)
        dFRR = lFP/(lFP + lTN)

        # fill test object
        dicTest = dicAdditionalInformation
        dicTest["challenge"] = sChallengeName
        dicTest["errorrate"] = dErrorRate
        dicTest["TP"] = lTP
        dicTest["TN"] = lTN
        dicTest["FP"] = lFP
        dicTest["FN"] = lFN
        dicTest["accuracy"] = dAccuracy
        dicTest["recall"] = dRecall
        dicTest["precision"] = dPrecision
        dicTest["F1_score"] = dF1score
        dicTest["FAR"] = dFAR
        dicTest["FRR"] = dFRR

        # save test in db
        if autosave_to_db:
            self.__save_test(dicTest)

        return dicTest

    def __save_test(self, dicTest):
        """ saves a test object to the database"""
        if not dicTest:
            raise Exception("Test object must not be None.")

        aTests = self._db.get(DB_TESTS_KEY, [])
        aTests.append(dicTest)
        self._db[DB_TESTS_KEY] = aTests
        self._db.commit()

    def save_test_threadsafe(self, dicTest, lock):
        """ saves a test object to the database threadsafe"""
        lock.acquire()
        self.__save_test(dicTest)
        lock.release()

    def get_tests(self):
        """getting all tests

        Returns:
            :obj:`list` of :obj:: `obj`:  List of all tests executed
        """
        return self._db.get(DB_TESTS_KEY, [])

    def clear_tests(self):
        """ delete all tests from the database """
        self._db[DB_TESTS_KEY] = []
        self._db.commit()
